///////////////////////////////////////////////////
// CANmain.cpp
// 720020-1185x
//
// Applied Data Systems
//
// See ReadMe.txt for a description and history
//
///////////////////////////////////////////////////


//#define HARD_CODE	1
#define CAN_PORT    "CAN1:"
//#define CAN_PORT    "CAN2:"

// define to respond (normal operation)
#define SEND_MSG_REPLY


#include <windows.h>
#include "CANapp.h"

#define SENDER_MODE 1
#define RECEIVER_MODE 2

#define SID 1	// Sender ID
#define RID 2	// Receiver ID

#define TIMEOUT_MS 5000		// milliseconds sender should wait for response

// Event name for the CAN event.  Could be read from the registry.
#define CAN_EVENT_NAME TEXT("EV_APP0")

BOOL SenderThread(LPVOID p);
BOOL ReceiverThread(LPVOID p);
BOOL gVerbose = FALSE;
ULONG nMsgCycles = 0;

// WinMain
//
// Main application thread.  Opens the CAN port and initializes CAN controller
// settings.  Spawns either the sender or receiver thread, and then processes 
// user input until user chooses to quit.  When the user quits, WinMain 
// terminates the running thread and closes the CAN port.

int WINAPI WinMain(
    HINSTANCE   hInst,
    HINSTANCE   hInstPrev,
    LPTSTR      lpszCmdLine,
    int         nCmdShow)
{
	int nMode;
	HANDLE hCanPort;
	WCHAR wszPortName [] = TEXT(CAN_PORT);	// Can port name
	HANDLE hCanThread;
	ULONG nBaud;
	char cKey;
	CAN_MSG_FILTER AFilter;
	
	// Open the CAN port and get the port handle
	hCanPort = CreateFile(	wszPortName, 
							GENERIC_WRITE | GENERIC_READ, 
							0, 
							NULL, 
							OPEN_EXISTING, 
							FILE_ATTRIBUTE_NORMAL, 
							NULL);
	if (hCanPort == NULL) {
		wsprintf(TEXT("Could not open port %s.\r\n"), wszPortName);
		printf("GetLastError returned 0x%x", GetLastError());
		return 1;
	} else wprintf(TEXT("Opened %s.\r\n"), wszPortName);
	RETAILMSG(1, (TEXT("Opened %s\r\n"), wszPortName));

	// Get the driver version
	TCHAR version[64];


	if (!DeviceIoControl(hCanPort, 
						 IOCTL_GET_DRIVER_VERSION, 
						 NULL,
						 0,
						 &version,
						 64,
						 NULL,
						 NULL))
	{
		printf("DeviceIoControl IOCTL_GET_DRIVER_VERSION failed.\r\n");
		printf("Version is likely < 1.2 and IOCTL is unsupported\r\n");
	} 
	else
	{
		// because printf has a bug with decimal strings we have to go char by char
		printf("Driver version: %c%c%c\r\n", version[0], version[1], version[2]);
	}
	

	// Reset the CAN chip
	if (!DeviceIoControl(hCanPort, IOCTL_RESET_CHIP, NULL, 0, NULL, 0, 
		NULL, NULL)) {
		printf("DeviceIoControl IOCTL_RESET_CHIP failed.\r\n");
		printf("GetLastError returned 0x%x", GetLastError());
		return 1;
	} else printf("Reset CAN chip.\r\n");
	RETAILMSG(1, (TEXT("Reset CAN chip.\r\n")));

	
	// Prompt user to select a baudrate
	nBaud = 0;
	while (!nBaud) {
		printf("<A> 100 kbit/s\r\n<B> 500 kbit/s\r\n<C> 1000 kbit/s\r\n");
		printf("Select a baudrate for the CAN bus:\r\n");
#ifndef HARD_CODE
		scanf("%s", &cKey);
#else
		cKey = 'B';   // hard code to 500K
#endif

		switch (cKey) {
			case 'a':
			case 'A':
				nBaud = 100;
				break;
			case 'b':
			case 'B':
				nBaud = 500;
				break;
			case 'c':
			case 'C':
				nBaud = 1000;
				break;
			default:
				break;
		}
	}
	
	// Set the baudrate
	if (!DeviceIoControl(hCanPort, IOCTL_CAN_SET_BAUDRATE, &nBaud, 
		sizeof(ULONG), NULL, 0, NULL, NULL)) {
		printf("DeviceIoControl IOCTL_CAN_SET_BAUDRATE failed.\r\n");
		printf("GetLastError returned 0x%x", GetLastError());
		return 1;
	} else printf("Baudrate set to %d kbit/s.\r\n", nBaud);
	
	
	// Prompt user to make this device either the sender or receiver
	nMode = 0;
	while (!nMode) {
		printf("Would you like this device to be the message sender or receiver?\r\n");
		printf("Type <S> for sender or <R> for Receiver: ");
#ifndef HARD_CODE
		scanf("%s", &cKey);
#else
		cKey = 'S';  // hard code to sender
#endif

		switch (cKey) {
			case 's':
			case 'S':
				nMode = SENDER_MODE;
				break;
			case 'r':
			case 'R':
				nMode = RECEIVER_MODE;
				break;
			default:
				break;
		}
	}

	// Adjust acceptance filter settings to only accept messages addressed to this device (ID = nMode)
	// Now we don't have to check the ID field in software.
	AFilter.mode = 1;					// Use single filter mode
	AFilter.mask0 = 0x00;				// ID bits 3-11 will be checked
	AFilter.code0 = 0x00;				// ID bits 3-11 must all be 0
	AFilter.mask1 = 0x0F;				// ID bits 0-2 and RTR bit will be checked
	AFilter.code1 = (nMode << 5);		// Only accept messages with ID bits 0-1 = nMode
	AFilter.code1 &= ~(0x10);			// Do not accept messages with RTR bit set
	AFilter.mask2 = 0xFF;				// Don't care about data byte 1
	AFilter.mask3 = 0xFF;				// Don't care about data byte 2

	// Apply acceptance filter settings
	if (!DeviceIoControl(	hCanPort, IOCTL_CAN_SET_ACCEPTANCE_FILTER, &AFilter, sizeof(CAN_MSG_FILTER),
							NULL, 0, NULL, NULL)) {
		printf("DeviceIoControl IOCTL_CAN_SET_ACCEPTANCE_FILTER failed.\r\n");
		printf("GetLastError returned 0x%x", GetLastError());
		return 1;
	}


	// Clear the CAN queue.  Get rid of any old messages before we start.
	if (!DeviceIoControl(hCanPort, IOCTL_CAN_CLEAR_QUEUE, NULL, 0, NULL, 0, NULL, NULL)) {
		printf("DeviceIoControl IOCTL_CAN_CLEAR_QUEUE failed.\r\n");
		printf("GetLastError returned 0x%x", GetLastError());
		return 1;
	} else printf("Cleared CAN queue.\r\n");
	RETAILMSG(1, (TEXT("Cleared CAN queue.\r\n")));


	// Create either the sender or receiver thread.
	if (nMode == SENDER_MODE) {	// Device will be the sender
		hCanThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)SenderThread, &hCanPort, 0, NULL);
	}
	else { 			//bMode == RECEIVE_MODE, device will be the receiver
		hCanThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReceiverThread, &hCanPort, 0, NULL);
	}

	// Prompt the user to press q to quit, or v to toggle verbose mode.
	// If the device is the sender, selecting a, b, c, or d sets the number of messages to send.
	// (For the receiver, setting nMsgCycles has no effect)
	do {
		if (nMode == RECEIVER_MODE) {
			printf("Type <Q> to quit or <V> to toggle verbose mode: ");
		}
#ifndef HARD_CODE
		scanf("%s", &cKey);
#else
		cKey = 'V';		// hard code verbose
		nMsgCycles = 5;   // use just 5 cycles
#endif

		switch (cKey) {
			case 'a':
			case 'A': 	nMsgCycles += 1;
						break;
			case 'b':
			case 'B': 	nMsgCycles += 10;
						break;
			case 'c':
			case 'C': 	nMsgCycles += 100;
						break;
			case 'd':
			case 'D': 	nMsgCycles += 1000;
						break;
			case 'v':
			case 'V':	gVerbose = !gVerbose;
						if (gVerbose) printf("VERBOSE MODE ON\r\n");
						else printf("VERBOSE MODE OFF\r\n");
			default:	break;
		}
		Sleep(1000);

#ifdef HARD_CODE
		while (nMsgCycles > 0)
			Sleep(500);

		cKey = 'Q';       // exit when done
		Sleep(1000);
#endif

	} while ((cKey != 'q') && (cKey != 'Q'));
	
	// User has pressed 'q' to quit.
	TerminateThread(hCanThread, 0);		// End the sender or receiver thread
	CloseHandle(hCanPort);				// Close the CAN port
	return 0;
}


// SenderThread
//
// When the user sets nMsgCycles, the sender begins by transmitting a CAN message
// to the receiver containing an 8 byte LONGLONG integer.  The sender then waits
// MS_TIMEOUT milliseconds for a response to be sent back from the receiver.  It
// expects the response to contain the integer it sent, plus one.  If a different
// response is received, an error message is displayed.  If no response is
// received within the time allowed, the sender retransmits the same CAN message.
// The sender counts the total number of messages successfully transmitted.
// When all message send-receive cycles have been completed, the sender calculates
// and displays throughput statistics.
//
// Parameters: 
// LPVOID p - pointer to the CAN port handle 

BOOL SenderThread(LPVOID p) {
	CAN_MSG TxCanMsg, RxCanMsg;			// Transmit and receive message structs
	DWORD dRet;							// Return value for WaitForSingleObject
	HANDLE hCanPort = *((HANDLE *) p);	// Get port handle from thread params
	HANDLE hCanAppEvent;				// Handle to event notifying message received

	DWORD dwBytesWritten, dwBytesRead;
	UINT nNumMessages;

	// Statistics
	double dDataThroughput;
	double dRawThroughput;
	ULONG nBytesXfer = 0;		// Number of data bytes transfered
	ULONG nMessagesXfer = 0;	// Number of messages transfered
	DWORD dwStartTime = 0;		// Initial millisecond count
	DWORD dwEndTime = 0;

	// Format transmit message
	TxCanMsg.length = 8; 				// 8 bytes in data buffer
	TxCanMsg.id = RID;					// Message recipient ID
	TxCanMsg.lData = 0;					// Message data, cast as LONG LONG (64 bits = 8 bytes)
	TxCanMsg.flags = 0;					// No flags needed

	// Get handle to the CAN event to notify us when we receive a CAN message
	hCanAppEvent = CreateEvent(NULL, FALSE, FALSE, CAN_EVENT_NAME);

	printf("Each message cycle is one roundtrip from sender to receiver and back to sender.\r\n");

	while (1) {
		if (nMsgCycles == 0) {
			printf("<A> 1 cycle   <B> 10 cycles   <C> 100 cycles   <D> 1000 cycles\r\n");
			printf("Select number of cycles to run, type <Q> to quit or <V> to toggle verbose mode:");
		}
		while (nMsgCycles == 0) Sleep(1000); // Wait for user to set nMsgCycles (in WinMain())
		if (nMsgCycles > 0) {
			// Intialize statistics information
			nBytesXfer = 0;
			nMessagesXfer = 0;
			dwStartTime = GetTickCount();		// Get initial millisecond timestamp
			while (nMsgCycles > 0) {
				// Send a message to the receiver.
				if (gVerbose)
					printf("Writing CAN message: data = %ld, id = %ld\r\n", TxCanMsg.lData, TxCanMsg.id);
				if (!WriteFile(hCanPort, &TxCanMsg, sizeof(CAN_MSG), &dwBytesWritten, NULL)) {
					RETAILMSG(1, (TEXT("\r\nERROR: Could not write to CAN port.\r\n")));
					return FALSE;
				}
				
				// Update transfer stats
				nBytesXfer += 8;
				nMessagesXfer++;
				
				// Wait until a message is received or TIMEOUT_MS milliseconds pass
				dRet = WaitForSingleObject(hCanAppEvent, TIMEOUT_MS);
				if (dRet == WAIT_OBJECT_0) {	// A message was received
					if (gVerbose)
						wprintf(L"Got event: " CAN_EVENT_NAME L".\r\n");
					// Read messages from the queue until it is empty
					do {
						if (!ReadFile(hCanPort, &RxCanMsg, sizeof(CAN_MSG), &dwBytesRead, NULL)) {
							RETAILMSG(1, (TEXT("\r\nERROR: Could not read from CAN port.\r\n")));
							return FALSE;
						}
						if (gVerbose)
							printf("Received CAN message: data = %ld, id = %ld\r\n", RxCanMsg.lData, RxCanMsg.id);
						// If there are still messages left in queue then this message is old
						// If so, discard it and read the next one.
						if (!DeviceIoControl(hCanPort, IOCTL_CAN_GET_NUM_MSGS, NULL, 0, &nNumMessages, sizeof(UINT), NULL, NULL)) {
							RETAILMSG(1, (TEXT("\r\nERROR: DeviceIoControl code IOCTL_CAN_GET_MSGS failed.\r\n")));
							return FALSE;
						}
						if (nNumMessages > 0) {
							printf("Found old data = %ld in RX FIFO.", RxCanMsg.lData);
							printf("Reading next message.\r\n");
						}
					} while (nNumMessages > 0);
					// Verify that received message contains correct data
					if (RxCanMsg.lData != (TxCanMsg.lData + 1))
						printf("Data error: Sent %ld, but received %ld.  Expected %ld\r\n",
							TxCanMsg.lData, RxCanMsg.lData,	(TxCanMsg.lData + 1) ); 
				} else if (dRet == WAIT_TIMEOUT) {	// Timeout limit was reached
					printf("Timed out waiting for response.  Resending...\r\n");
					// Resend the last message
					if (gVerbose)
						printf("Writing CAN message: data = %ld, id = %ld\r\n", TxCanMsg.lData, TxCanMsg.id);
					
					if (!WriteFile(hCanPort, &TxCanMsg, sizeof(CAN_MSG), &dwBytesWritten, NULL)) {
						RETAILMSG(1, (TEXT("\r\nERROR: Could not write to CAN port.\r\n")));
						return FALSE;
					}
					// don't increment transmit byte count since no new bytes were transmitted
				}
				
				// Update transfer statistics
				nBytesXfer += 8;
				nMessagesXfer++;
				TxCanMsg.lData += 2;	// Prepare new message data to send
				nMsgCycles--;
			}
			dwEndTime = GetTickCount();			// Get current time in milliseconds
			// Calculate throughput stats
			dDataThroughput = ((double) nBytesXfer) / ((double) (dwEndTime - dwStartTime));
			dRawThroughput = ((double) nMessagesXfer) / ((double) (dwEndTime - dwStartTime));
			// Display stats
			printf("Messages transfered: %ld messages (%ld bits)\r\n", nMessagesXfer, nMessagesXfer * 106);
			printf("Data transfered: %ld bytes (%ld bits)\r\n", nBytesXfer, nBytesXfer * 8);
			printf("Time elapsed: %ld ms\r\n", dwEndTime - dwStartTime);
			printf("Total throughput: %lg messages/s (%lg kbit/s)\r\n", dRawThroughput * 1000, dRawThroughput *106);
			printf("Data throughput: %lg kB/s (%lg kbit/s)\r\n\n", dDataThroughput, dDataThroughput * 8);
		}
		Sleep(1000);
	}
	
	return 0;
}


// ReceiverThread
//
// Waits for a CAN message.  Reads the message data as a LONGLONG integer and
// adds 1 to it.  Sends a message with the new data value back to the sender.
// Loops until terminated by main application thread.
//
// Parameters: 
// LPVOID p - pointer to the CAN port handle 

BOOL ReceiverThread(LPVOID p) {
	CAN_MSG TxCanMsg, RxCanMsg;			// Transmit and receive message structs
	DWORD dRet;							// Return value for WaitForSingleObject
	HANDLE hCanPort = *((HANDLE *) p);	// Get port handle from thread params
	HANDLE hCanAppEvent;				// Handle to event notifying message received
	
	DWORD dwBytesWritten, dwBytesRead;
	UINT nNumMessages;


	int nCount = 0;
	
	// Format the message the receiver sends
	TxCanMsg.length = 8; 				// 8 bytes in data buffer
	TxCanMsg.id = SID;					// Message recipient ID is the sender ID
	TxCanMsg.lData = 0;					// Message data, cast as ULONG (64 bits = 8 bytes)
	TxCanMsg.flags = 0;					// No flags needed
	
	// Get handle to the CAN event to notify us when we receive a CAN message
	hCanAppEvent = CreateEvent(NULL, FALSE, FALSE, CAN_EVENT_NAME);
	
	printf("\r\nReceiver thread has started.\r\n");

	while (1) {

		// Wait infinately until a message is received
		dRet = WaitForSingleObject(hCanAppEvent, INFINITE);
		if (gVerbose)
			wprintf(L"Got event: " CAN_EVENT_NAME L".\r\n");
		do {
			// Read messages from the queue until it is empty
			if (!ReadFile(hCanPort, &RxCanMsg, sizeof(CAN_MSG), &dwBytesRead, NULL)) {
				RETAILMSG(1, (TEXT("\r\nERROR: Could not read from CAN port\r\n")));
				return FALSE;
			}

			nCount++; 

			if (gVerbose)
				printf("Received CAN message:#%d, data = %lx, id = %d\r\n", nCount, RxCanMsg.lData, RxCanMsg.id);
			// If there are still messages left in queue then this message is old
			// If so, discard it and read the next one.
			if (!DeviceIoControl(hCanPort, IOCTL_CAN_GET_NUM_MSGS, NULL, 0, &nNumMessages, sizeof(UINT), NULL, NULL)) {
				RETAILMSG(1, (TEXT("\r\nERROR: DeviceIoControl code IOCTL_CAN_GET_MSGS failed.\r\n")));
				return FALSE;
			}
			if (nNumMessages > 0) {
				printf("Found old data = %ld in RX FIFO.  ", RxCanMsg.lData);
				printf("Reading next message.\r\n");
			}
		} while (nNumMessages > 0);
#ifdef SEND_MSG_REPLY
		// Modify the received message and send it back to the sender
		TxCanMsg.lData = (RxCanMsg.lData + 1);
		if (gVerbose)
			printf("Writing CAN message: data = %ld, id = %ld\r\n", TxCanMsg.lData, TxCanMsg.id);
		if (!WriteFile(hCanPort, &TxCanMsg, sizeof(CAN_MSG), &dwBytesWritten, NULL)) {
			RETAILMSG(1, (TEXT("\r\nERROR: Could not write to CAN port\r\n")));
			return FALSE;
		}
#endif
	}
	return TRUE;
}
